from typing import Union, Optional
import os
import pandas as pd

from qiskit_metal import Dict, draw
from qiskit_metal.renderers.renderer_base import QRendererAnalysis
from qiskit_metal.renderers.renderer_gmsh.gmsh_renderer import QGmshRenderer
from qiskit_metal.renderers.renderer_elmer.elmer_runner import ElmerRunner


def load_capacitance_matrix_from_file(filename: str) -> pd.DataFrame:
    """Loads capacitance matrix from file.

    Args:
        filename (str): A '.txt' file containing the capacitance matrix.

    Returns:
        pd.DataFrame:: Table containing capacitance matrix.
    """
    return pd.read_csv(filename, delimiter=' ', index_col=0)


class QElmerRenderer(QRendererAnalysis):
    """Extends QRendererAnalysis class and imports meshes from Gmsh using the ElmerFEM python API.

    QElmerRenderer Default Options:
        * simulation_type -- Type of ElmerFEM simulation to run.
                                Default options found in elmer_configs.py under "simulations".
        * simulation_dir -- Directory to store simulation-related files.
        * mesh_file -- Name of GMSH output mesh file ending in '.msh'.
        * simulation_input_file -- Name of ElmerFEM simulation input file ending in '.sif'.
        * postprocessing_file -- Name of ElmerFEM postprocessing output file ending in '.msh'.
        * output_file -- Name of ElmerFEM results output file ending in '.result'.

    QElmerRenderer Default Setup:
        * capacitance -- Default setup for capacitance simulation:
            * Calculate_Electric_Field -- Defaults to True.
            * Calculate_Electric_Energy -- Defaults to True.
            * Calculate_Capacitance_Matrix -- Defaults to True.
            * Capacitance_Matrix_Filename -- Defaults to "cap_matrix.txt".
            * Linear_System_Solver -- Defaults to "Iterative".
            * Steady_State_Convergence_Tolerance -- Defaults to 1.0e-5.
            * Nonlinear_System_Convergence_Tolerance -- Defaults to 1.0e-7.
            * Nonlinear_System_Max_Iterations -- Defaults to 20.
            * Linear_System_Convergence_Tolerance -- Defaults to 1.0e-10.
            * Linear_System_Max_Iterations -- Defaults to 500.
            * Linear_System_Iterative_Method -- Defaults to "BiCGStab".
            * BiCGstabl_polynomial_degree -- Defaults to 2.

        * eigenmode -- Default setup for Eigenmode simulation (to be added in future).

        * materials -- Materials used in the simulation. Defaults to `["vacuum", "silicon"]`.

        * constants -- Default constants to be included in the simulation:
            * Permittivity_of_Vacuum -- Defaults to 8.8542e-12 F/m.
            * Unit_Charge -- Defaults to 1.602e-19 C.

    QGmshRenderer Default Options:
        # Buffer between max/min x and edge of ground plane, in mm
        * x_buffer_width_mm -- Buffer between max/min x and edge of ground plane, in mm
        * y_buffer_width_mm -- Buffer between max/min y and edge of ground plane, in mm
        * mesh -- to define meshing parameters
            * max_size -- upper bound for the size of mesh node
            * min_size -- lower bound for the size of mesh node
            * max_size_jj -- maximum size of mesh nodes at jj
            * smoothing -- mesh smoothing value
            * nodes_per_2pi_curve -- number of nodes for every 2π radians of curvature
            * algorithm_3d -- value to indicate meshing algorithm used by Gmsh
            * num_threads -- number of threads for parallel meshing
            * mesh_size_fields -- specify mesh size field parameters
                * min_distance_from_edges -- min distance for mesh gradient generation
                * max_distance_from_edges -- max distance for mesh gradient generation
                * distance_delta -- delta change in distance with each consecutive step
                * gradient_delta -- delta change in gradient with each consecutive step
        * colors -- specify colors for the mesh elements, chips or layers
            * metal -- color for metallized entities
            * jj -- color for JJs
            * dielectric -- color for dielectric entity
    """

    default_options = Dict(
        simulation_type="steady_3D",
        simulation_dir="./simdata",
        mesh_file="out.msh",
        simulation_input_file="case.sif",
        postprocessing_file="case.msh",
        output_file="case.result",
    )

    default_setup = Dict(
        capacitance=Dict(
            Calculate_Electric_Field=True,
            Calculate_Electric_Energy=True,
            Calculate_Capacitance_Matrix=True,
            Capacitance_Matrix_Filename="cap_matrix.txt",
            Linear_System_Solver="Iterative",
            Steady_State_Convergence_Tolerance=1.0e-5,
            Nonlinear_System_Convergence_Tolerance=1.0e-7,
            Nonlinear_System_Max_Iterations=20,
            Linear_System_Convergence_Tolerance=1.0e-10,
            Linear_System_Max_Iterations=500,
            Linear_System_Iterative_Method="BiCGStab",
            BiCGstabl_polynomial_degree=2,
        ),

        # TODO: Update this later when working on eigenmode setup
        eigenmode=Dict(),
        materials=["vacuum", "silicon"],
        constants=Dict(
            Permittivity_of_Vacuum=8.8542e-12,
            Unit_Charge=1.602e-19,
        ),
    )

    name = "elmer"
    """Name"""

    def __init__(self,
                 design: 'MultiPlanar',
                 layer_types: Union[dict, None] = None,
                 initiate=True,
                 options: Dict = None):
        """
        Args:
            design ('MultiPlanar'): The design.
            layer_types (Union[dict, None]): the type of layer in the format:
                                                dict(metal=[...], dielectric=[...]).
                                                Defaults to None.
            initiate (bool): True to initiate the renderer (Default: False).
            settings (Dict, optional): Used to override default settings. Defaults to None.
        """
        default_layer_types = dict(metal=[1], dielectric=[3])
        self.layer_types = default_layer_types if layer_types is None else layer_types

        super().__init__(design=design, initiate=initiate, options=options)

    @property
    def initialized(self):
        """Abstract method. Must be implemented by the subclass.
            Is renderer ready to be used?
            Implementation must return boolean True if successful. False otherwise.
        """
        return (hasattr(self, "_elmer_runner") and self.gmsh.initialized)

    def initialize_renderer(self):
        """Initializes the Gmsh and Elmer renderer.
        NOTE TO USER: this should be used when using
        the QElmerRenderer through design.renderers.elmer instance.
        
        Example: To change one of the options exposed by QGmshRenderer,
        one has to first initiate the QElmerRenderer using this method
        and then change the property like so,

        ```
        design.renderers.elmer.initialize_renderer()
        design.renderers.elmer.gmsh.options.mesh.num_threads = 1
        ...
        ```
        """
        self.gmsh = QGmshRenderer(self.design, self.layer_types)
        self._elmer_runner = ElmerRunner()

    def _initiate_renderer(self):
        """Initializes the Gmsh and Elmer renderer.
        NOTE: This is automatically called when the USER
        specifically imports QElmerRenderer in a jupyter notebook
        and instantiates it."""
        self.gmsh = QGmshRenderer(self.design, self.layer_types)
        self._elmer_runner = ElmerRunner()
        return True

    def _close_renderer(self):
        """Finalizes the Gmsh renderer"""
        self.gmsh.close()

    def close(self):
        """Public method to close the Gmsh renderer"""
        return self._close_renderer()

    def render_design(
        self,
        selection: Union[list, None] = None,
        open_pins: Union[list, None] = None,
        box_plus_buffer: bool = True,
        draw_sample_holder: bool = True,
        skip_junctions: bool = False,
        mesh_geoms: bool = True,
        ignore_metal_volume: bool = False,
        omit_ground_for_layers: Optional[list[int]] = None,
    ):
        """Render the design in Gmsh and apply changes to modify the geometries
        according to the type of simulation. Simulation parameters provided by the user.

        Args:
            selection (Union[list, None], optional): List of selected components
                                                        to render. Defaults to None.
            open_pins (Union[list, None], optional): List of open pins that are open.
                                                        Defaults to None.
            box_plus_buffer (bool, optional): Set to True for adding buffer to
                                                        chip dimensions. Defaults to True.
            draw_sample_holder (bool, optional): To draw the sample holder box. Defaults to True.
            skip_junctions (bool, optional): Set to True to sip rendering the
                                                        junctions. Defaults to False.
            mesh_geoms (bool, optional): Set to True for meshing the geometries.
                                                        Defaults to True.
            ignore_metal_volume (bool, optional): ignore the volume of metals and replace
                                                        it with a list of surfaces instead.
                                                        Defaults to False.
            omit_ground_for_layers (Optional[list[int]]): omit rendering the ground plane for
                                                         specified layers. Defaults to None.
        """

        # For handling the case when the user wants to use
        # QElmerRenderer from design.renderers.elmer instance.
        if not self.initialized:
            self._initiate_renderer()

        self.gmsh.render_design(selection=selection,
                                open_pins=open_pins,
                                box_plus_buffer=box_plus_buffer,
                                draw_sample_holder=draw_sample_holder,
                                skip_junctions=skip_junctions,
                                mesh_geoms=mesh_geoms,
                                ignore_metal_volume=ignore_metal_volume,
                                omit_ground_for_layers=omit_ground_for_layers)

        self.qcomp_geom_table = self.get_qgeometry_table()
        self.nets = self.assign_nets(open_pins=open_pins)

    def get_qgeometry_table(self) -> pd.DataFrame:
        """Combines the "path" and "poly" qgeometry tables into a single table, and adds
                column containing the minimum z coordinate of the layer associated with
                each qgeometry.

        Raises:
            ValueError: Raised when component in selection of qcomponents is not in QDesign.

        Returns:
            pd.DataFrame: Table with elements in both "path" and "poly" qgeometry tables.
        """

        if self.gmsh.case == 0:
            qcomp_ids = self.gmsh.qcomp_ids
        elif self.gmsh.case == 1:
            qcomp_ids = list(set(self.design._components.keys()))
        elif self.gmsh.case == 2:
            raise ValueError("Selection provided is invalid.")

        metal_layers = self.layer_types["metal"]

        mask = lambda table: table["component"].isin(qcomp_ids) & ~table[
            "subtract"] & table["layer"].isin(metal_layers)

        min_z = lambda layer: min(
            sum(self.gmsh.get_thickness_zcoord_for_layer_datatype(layer)),
            self.gmsh.get_thickness_zcoord_for_layer_datatype(layer)[1])

        path_table = self.design.qgeometry.tables["path"]
        poly_table = self.design.qgeometry.tables["poly"]
        qcomp_paths = path_table[mask(table=path_table)]
        qcomp_polys = poly_table[mask(table=poly_table)]
        qcomp_geom_table = pd.concat([qcomp_paths, qcomp_polys],
                                     ignore_index=True)

        qcomp_geom_table['min_z'] = qcomp_geom_table['layer'].apply(min_z)
        qcomp_geom_table.sort_values(by=['min_z'],
                                     inplace=True,
                                     ignore_index=True)

        return qcomp_geom_table

    def assign_nets(
        self,
        open_pins: Union[list,
                         None] = None) -> dict[Union[str, int], list[str]]:
        """Assigns a netlist number to each galvanically connected metal region,
        and returns a dictionary with each net as a key, and the corresponding list of
        geometries associated with that net as values.

        Args:
            open_pins (Union[list, None], optional): List of tuples of pins that are open.
                                                        Defaults to None.

        Returns:
            dict[Union[str, int], list[str]]: dictionary with keys for each net, and list of
                    values with the corresponding geometries associated with that net as values.
        """

        netlists = dict()
        netlist_id = 0

        qgeom_names = self.qcomp_geom_table["name"]
        qcomp_names_for_qgeom = [
            list(self.design.components.keys())[i - 1]
            for i in self.qcomp_geom_table["component"]
        ]
        phys_grps = [
            s1 + '_' + s2 for s1, s2 in zip(qcomp_names_for_qgeom, qgeom_names)
        ]
        qgeom_idxs = list(range(len(self.qcomp_geom_table)))
        id_net_dict = {k: -1 for k in phys_grps}

        while len(qgeom_idxs) != 0:
            i = qgeom_idxs.pop(0)
            shape_i = self.qcomp_geom_table.iloc[[i]]["geometry"][i]
            chip_i = self.qcomp_geom_table.iloc[[i]]["chip"][i]
            layer_i = self.qcomp_geom_table.iloc[[i]]["layer"][i]
            thick_i, z_coord_i = self.gmsh.get_thickness_zcoord_for_layer_datatype(
                layer_i)
            id_net_dict[phys_grps[i]] = netlist_id if (
                id_net_dict[phys_grps[i]] == -1) else id_net_dict[phys_grps[i]]
            for j in qgeom_idxs:
                shape_j = self.qcomp_geom_table.iloc[[j]]["geometry"][j]
                chip_j = self.qcomp_geom_table.iloc[[j]]["chip"][j]
                layer_j = self.qcomp_geom_table.iloc[[j]]["layer"][j]
                thick_j, z_coord_j = self.gmsh.get_thickness_zcoord_for_layer_datatype(
                    layer_j)
                dist = shape_i.distance(shape_j)

                layers_touch = False
                if (layer_i == layer_j or z_coord_j == z_coord_i or
                        z_coord_i + thick_i == z_coord_j or
                        z_coord_j + thick_j == z_coord_i):
                    layers_touch = True

                if dist == 0.0 and chip_i == chip_j and layers_touch:
                    if id_net_dict[phys_grps[j]] == -1:
                        id_net_dict[phys_grps[j]] = id_net_dict[phys_grps[i]]
                    elif id_net_dict[phys_grps[j]] != id_net_dict[phys_grps[i]]:
                        net_id_i = id_net_dict[phys_grps[i]]
                        for k, v in id_net_dict.items():
                            if v == net_id_i:
                                id_net_dict[k] = id_net_dict[phys_grps[j]]

            if -1 not in id_net_dict.values():
                break

            netlist_id = max(list(id_net_dict.values())) + 1

        gnd_phys_grps = self.get_gnd_qgeoms(open_pins)
        gnd_netlist = list({
            net for (name, net) in id_net_dict.items()
            if (name in gnd_phys_grps)
        })

        netlists['gnd'] = list()
        for k, v in id_net_dict.items():
            if v in gnd_netlist:
                netlists['gnd'].append(k)
            else:
                if v not in netlists.keys():
                    netlists[v] = list()
                netlists[v].append(k)

        netlists = {
            (i - 1 if (k != 'gnd') else k): v
            for i, (k, v) in enumerate(netlists.items())
        }

        return netlists

    def get_gnd_qgeoms(self, open_pins: Union[list, None] = None) -> list[str]:
        """ Obtain a list of qgeometry names associated with pins shorted to ground.

        Args:
            open_pins (Union[list, None], optional): List of tuples of pins that are open.
                                                        Defaults to None.

        Returns:
            list[str]: Names of qgeometry components with pins connected to ground plane.
        """

        open_pins = open_pins if open_pins is not None else []
        qcomp_lst = self.design.components.keys()
        all_pins = list()
        gnd_qgeoms = set()

        for qcomp in qcomp_lst:
            qcomp_pins = self.design.components[qcomp].pins.keys()
            all_pins += list(zip([qcomp] * len(qcomp_pins), qcomp_pins))

        nets_table = self.design.net_info
        gnd_nets = list(nets_table[nets_table["pin_name"] == "short"]["net_id"])
        gnd_nets_mask = nets_table["net_id"].isin(gnd_nets)
        port_nets_table = nets_table[~gnd_nets_mask]
        port_pins_names = port_nets_table["pin_name"]
        qcomp_names_for_port_pins = [
            list(qcomp_lst)[i - 1] for i in port_nets_table["component_id"]
        ]
        port_pins = list(zip(qcomp_names_for_port_pins, port_pins_names))
        all_open_pins = list(set(port_pins + open_pins))
        gnd_pins = [pin for pin in all_pins if pin not in all_open_pins]

        for pin in gnd_pins:
            qcomp_name, qcomp_pin = pin
            pin_qcomp_id = self.design.components[qcomp_name].id

            if pin_qcomp_id in list(self.qcomp_geom_table["component"]):
                pin_qcomp_geom_table = self.qcomp_geom_table[
                    self.qcomp_geom_table["component"] == pin_qcomp_id]
                pin_point = draw.Point(
                    self.design.components[qcomp_name].pins[qcomp_pin]
                    ['middle'])

                for _, qgeom_row in pin_qcomp_geom_table.iterrows():
                    if pin_point.intersects(qgeom_row["geometry"]):
                        gnd_qgeoms.add(qcomp_name + '_' + qgeom_row["name"])

        return list(gnd_qgeoms)

    def run(self,
            sim_type: str,
            display_cap_matrix: bool = False) -> Union[None, pd.DataFrame]:
        """Runs ElmerFEM analysis.

        Args:
            sim_type (str): Type of simulation to be executed by ElmerFEM.
            display_cap_matrix (bool, optional): Whether or not to return cap matrix.
                                                    Defaults to False.

        Returns:
            Union[None, pd.DataFrame]: extracted capacitance matrix if
                            display_cap_matrix = True.
        """

        setup = self.default_setup[sim_type]
        sim_dir = self._options["simulation_dir"]
        meshfile = self._options["mesh_file"]
        sif_name = self._options["simulation_input_file"]
        if sim_type == "capacitance":
            cap_matrix_file = os.path.join(self._options["simulation_dir"],
                                           setup["Capacitance_Matrix_Filename"])

        self.logger.info("Running ElmerGrid on input mesh from Gmsh...")
        self._elmer_runner.run_elmergrid(sim_dir, meshfile)
        self.logger.info(f"Running ElmerSolver for solver type: '{sim_type}'")
        self._elmer_runner.run_elmersolver(sim_dir, sif_name)

        self.capacitance_matrix = self._get_capacitance_matrix(cap_matrix_file)
        if display_cap_matrix:
            return self.capacitance_matrix

    def save_capacitance_matrix(self, path: str):
        """Saves capacitance matrix to file.

        Args:
            path (str): path where capacitance matrix should be saved.
        """
        self.capacitance_matrix.to_csv(path, sep=' ', header=True)

    def _get_capacitance_matrix(self, filename: str) -> pd.DataFrame:
        """Reads SPICE capacitance matrix file generated by ElmerFEM, converts
        and returns Maxwell capacitance matrix

        Args:
            filename (str): Name of capacitance matrix file generated by ElmerFEM

        Returns:
             pd.DataFrame: Maxwell capacitance matrix
        """
        with open(filename, "r", encoding="utf-8") as f:
            lines = f.readlines()
            f.close()

        new_lines = [(l.replace("  ", " ").strip() + '\n') for l in lines]

        with open(filename, "w", encoding="utf-8") as f:
            f.writelines(new_lines)
            f.close()

        df = pd.read_csv(filename, delimiter=" ", header=None)

        # Convert from SPICE cap matrix to Maxwell cap matrix
        df2 = df.multiply(-1e15)
        row_sum = df2.sum(axis=0)
        for i, s in enumerate(row_sum):
            df2[i][i] = -s

        name_cap = {k: v[-1] for k, v in self.nets.items() if k != "gnd"}
        df2.rename(index=name_cap, columns=name_cap, inplace=True)

        gnd_caps = -1 * df2.sum(axis=0)
        df2.loc["ground_plane"] = gnd_caps
        df2["ground_plane"] = gnd_caps
        df2 = df2.multiply(
            self.default_setup["constants"]["Permittivity_of_Vacuum"])

        # dummy value set to high as it's shorted to ground
        df2["ground_plane"]["ground_plane"] = 300
        return df2

    def add_solution_setup(
        self,
        sim_name: str = "capacitance",
        solver_names: list[str] = ["capacitance", "postprocessing_gmsh"],
        equation_name: str = "poisson",
    ):
        """Initializes ElmerFEM analysis to run.
            Currently only supports Electrostatic capacitance extraction

        Args:
            sim_name (str, optional): Type of ElmerFEM analysis to initialize.
                                        Defaults to "capacitance".
            solver_names (list[str], optional): ElmerFEM solver to use. Defaults
                                                    to ["capacitance", "postprocessing_gmsh"].
            equation_name (str, optional): Type of equation for solver. Defaults to "poisson".
        """
        setup = self.default_setup
        cap_setup = {
            k.replace('_', ' '): v
            for k, v in self.default_setup[sim_name].items()
        }

        # Load simulation
        self._elmer_runner.load_simulation(self._options["simulation_type"])

        # Load constants
        constants = self.default_setup["constants"]
        for k in constants.keys():
            const_str = k.replace('_', ' ')
            self._elmer_runner.add_constant(const_str, 1)

        # Load solvers
        solvers = self._elmer_runner.load_solvers(solver_names=solver_names)
        self._elmer_runner.solver["capacitance"].update(cap_setup)

        # Load materials
        materials = self._elmer_runner.load_materials(setup["materials"])

        # Add equation
        eqn = self._elmer_runner.add_equation(equation_name, solvers=solvers)

        # Add bodies
        bodies = self.define_bodies(setup, eqn, materials)

        # Add boundary conditions
        boundaries, cap_body = self.define_boundaries()

        if cap_body == 1:
            self.logger.warning(
                "WARNING: No capacitance bodies added to the model other "
                "than those connected to ground. ElmerFEM cannot run a "
                "capacitance extraction analysis without any bodies. ")

        # Update capacitance bodies in solver
        self._elmer_runner.solver["capacitance"].update(
            {"Capacitance Bodies": cap_body - 1})

        # Write the simulation input file (SIF)
        self.write_sif()

    def define_bodies(self, setup: dict, equation: int,
                      materials: list[int]) -> list[list]:
        """Assigns bodies to simulation model.

        Args:
            setup (dict): Setup parameters to be used in simulation.
            equation (int): Type of equation for solver.
            materials (list[int]): Materials used by bodies in the design.

        Returns:
            list[list]: Information about the bodies in the design.
        """
        bodies = []
        for i, material in enumerate(setup["materials"]):
            if material == "vacuum":
                bodies += [
                    self._elmer_runner.add_body(
                        "vacuum_box",
                        [self.gmsh.physical_groups["global"]["vacuum_box"]],
                        equation, materials[i])
                ]
            if material == "silicon":
                substrate_geoms = dict()
                for layer, ph_geoms in self.gmsh.physical_groups.items():
                    if layer in ["global", "chips"]:
                        continue

                    sub = {
                        k: v
                        for k, v in ph_geoms.items()
                        if ("dielectric" in k) and ("_sfs" not in k)
                    }
                    substrate_geoms.update(sub)

                bodies += [
                    self._elmer_runner.add_body(name, [body], equation,
                                                materials[i])
                    for name, body in substrate_geoms.items()
                ]

        return bodies

    def define_boundaries(self) -> tuple[list[list], int]:
        """Assigns boundaries to simulation model.

        Returns:
            tuple[list[list], int]: Information about the bodies in the design,
                                        and total number of capacitance bodies.
        """

        boundaries = list()
        phys_grps_dict = dict()
        ground_planes = list()
        cap_body = 1
        for layer, ph_geoms in self.gmsh.physical_groups.items():
            if layer in ["global", "chips"]:
                continue

            for name, geom in ph_geoms.items():
                if ("ground_plane" in name) and ("sfs" in name):
                    ground_planes.append(geom)

                if "dielectric" in name:
                    continue

            phys_grps_dict.update(ph_geoms)

        for net, geom_names in self.nets.items():
            geom_id_list = [
                phys_grps_dict[f"{name}_sfs"] for name in geom_names
            ]

            if net == "gnd":
                ground_planes += geom_id_list
            else:
                boundaries += [
                    self._elmer_runner.add_boundary_condition(
                        geom_names[-1],
                        geom_id_list,
                        options={"Capacitance Body": cap_body})
                ]
                cap_body += 1

        self._elmer_runner.add_boundary_condition(
            "ground_plane", ground_planes, options={"Capacitance Body": 0})

        self._elmer_runner.add_boundary_condition(
            "FarField", [self.gmsh.physical_groups["global"]["vacuum_box_sfs"]],
            options={"Electric Infinity BC": True})

        return boundaries, cap_body

    def write_sif(self):
        """Write sif (simulation input file) used by ElmerFEM to run analysis"""
        filename = self._options["simulation_input_file"]
        sim_dir = self._options["simulation_dir"]
        self._elmer_runner.write_startinfo_and_sif(filename=filename,
                                                   sim_dir=sim_dir)

    def render_chips(self,
                     chips: Union[str, list[str]] = [],
                     draw_sample_holder: bool = True,
                     box_plus_buffer: bool = True):
        """Abstract method. Must be implemented by the subclass.
        Render all chips of the design.
        Calls render_chip for each chip.
        """
        pass

    def render_chip(self, chip_name: str):
        """Abstract method. Must be implemented by the subclass.
        Render the given chip.

        Args:
            name (str): chip to render
        """
        pass

    def render_layers(self,
                      draw_sample_holder: bool = True,
                      layers: Union[list[int], None] = None,
                      box_plus_buffer: bool = True):
        """Abstract method. Must be implemented by the subclass.
        Render all layers of the design.
        Calls render_layer for each layer.
        """
        self.gmsh.render_layers(draw_sample_holder, layers, box_plus_buffer)

    def render_layer(self, layer_number: str, datatype: int = 0):
        """Abstract method. Must be implemented by the subclass.
        Render the given layer.

        Args:
            name (str): layer to render
        """
        self.gmsh.render_layer(layer_number, datatype)

    def render_components(self, table_type: str):
        """Abstract method. Must be implemented by the subclass.
        Render all components of the design.
        If selection is none, then render all components.

        Args:
            selection (QComponent): Component to render.
        """
        self.gmsh.render_components(table_type)

    def render_component(self, component):
        """Abstract method. Must be implemented by the subclass.
        Render the specified component.

        Args:
            component (QComponent): Component to render.
        """
        self.gmsh.render_component(component)

    def render_element(self, qgeom: pd.Series, table_type: str):
        """Abstract method. Must be implemented by the subclass.
        Render the specified element

        Args:
            element (Element): Element to render.
        """
        self.gmsh.render_element(qgeom, table_type)

    def render_element_path(self, path: pd.Series):
        """Abstract method. Must be implemented by the subclass.
        Render an element path.

        Args:
            path (str): Path to render.
        """
        self.gmsh.render_element_path(path)

    def render_element_junction(self, junc: pd.Series):
        """Abstract method. Must be implemented by the subclass.
        Render an element junction.

        Args:
            junc (str): Junction to render.
        """
        self.gmsh.render_element_junction(junc)

    def render_element_poly(self, poly: pd.Series):
        """Abstract method. Must be implemented by the subclass.
        Render an element poly.

        Args:
            poly (Poly): Poly to render.
        """
        self.gmsh.render_element_poly(poly)

    def save_screenshot(self, path: str = None, show: bool = True):
        """Save the screenshot.

        Args:
            path (str, optional): Path to save location.  Defaults to None.
            show (bool, optional): Whether or not to display the screenshot.  Defaults to True.

        Returns:
            pathlib.WindowsPath: path to png formatted screenshot.
        """
        self.gmsh.save_screenshot(path, show)

    def launch_gmsh_gui(self):
        """Launch Gmsh GUI for viewing the model.
        """
        self.gmsh.launch_gui()

    def export_mesh(self):
        """Export Gmsh mesh
        """
        self.gmsh.export_mesh(self._options["mesh_file"])

    def display_post_processing_data(self):
        """Import data given by ElmerFEM for Post-Processing in Gmsh
        """
        postprocessing_file = os.path.join(self._options["simulation_dir"],
                                           self._options["postprocessing_file"])
        self.gmsh.import_post_processing_data(postprocessing_file)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from PySide6.QtWidgets import (QAbstractItemView, QFileDialog, QMainWindow,
                               QMessageBox)

from .list_model_base import DynamicList
from .renderer_gds_model import RendererGDS_Model
from .renderer_gds_ui import Ui_MainWindow


class RendererGDSWidget(QMainWindow):
    """Contains methods associated with GDS Renderer button."""

    def __init__(self, parent: 'QMainWindow', gui: 'MetalGUI'):
        """Get access to design, which has the components. Then set up the
        model and view.

        Args:
            parent (QMainWindow): The parent window
            gui (MetalGUI): The metal GUI
        """
        super().__init__(parent)

        # Access design and Metal GUI:
        self._gui = gui

        # Use UI template from Qt Designer:
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        # Set up models for component list and options tree:
        self.listView = self.ui.listView
        self.list_model = DynamicList(self.design)
        self.listView.setModel(self.list_model)

        self.tree_model = RendererGDS_Model(self, gui, self.ui.treeView)
        self.ui.treeView.setModel(self.tree_model)
        self.ui.treeView.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.ui.treeView.setHorizontalScrollMode(
            QAbstractItemView.ScrollPerPixel)

    def set_design(self, new_design: 'QDesign'):
        """Swaps out reference to design, which changes the reference to the
        dictionary.

        Args:
            new_design (QDesign): The design
        """
        self.list_model.update_src(self.design)

    @property
    def design(self):
        """Returns the design."""
        return self._gui.design

    def refresh(self):
        """Refreshes list of components."""
        self.list_model.populate_list()

    def select_all(self):
        """Marks all components for export."""
        self.list_model.select_all()

    def deselect_all(self):
        """Empties list of components to export."""
        self.list_model.deselect_all()

    def get_checked(self) -> list:
        """Gets list of all selected components to export.

        Returns:
            list: List of selected components
        """
        return self.list_model.get_checked()

    def browse_folders(self):
        """Browses available folders in system."""
        destination_folder = QFileDialog.getSaveFileName(
            None,
            'Select a new location to export to',
            'my_qdesign.gds',
            selectedFilter='*.gds')[0]
        self.ui.lineEdit.setText(destination_folder)

    def export_file(self):
        """Renders a subset or all of the components in design and exports it.

        If the list of components to export is smaller than the total
        number of components, highlight_qcomponents is included as an
        argument. Otherwise it is not.
        """
        filename = self.ui.lineEdit.text()
        components_to_export = self.get_checked()
        a_gds = self.design.renderers.gds
        if filename and components_to_export:
            if len(components_to_export) == len(self.design.components):
                a_gds.export_to_gds(filename)
            else:
                a_gds.export_to_gds(filename,
                                    highlight_qcomponents=components_to_export)
            self.close()
        else:
            QMessageBox.warning(
                self, "Error",
                "Please enter a valid file name and \n select at least one component."
            )



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Pan and zoom interaction to plug on a matplotlib Figure.

Interaction:
    - Zoom in/out with the mouse wheel
    - Pan figures by dragging the mouse with left button pressed
    - Select a zoom-in area by dragging the mouse with right button pressed

It provides a figure_pz function to create a Figure with interaction.

Example:

.. code-block:: python

    import matplotlib.pyplot as plt
    from mpl_interaction import figure_pz
    fig = figure_pz()
    ax = fig.add_subplot(1, 1, 1)
    ax.plot((1, 2, 1))
    plt.show()

The :class:`PanAndZoom` class can be used to add interaction
to an existing Figure.

Example:

.. code-block:: python

    import matplotlib.pyplot as plt
    from mpl_interaction import PanAndZoom
    fig = plt.figure()
    pan_zoom = PanAndZoom(fig)  # Add support for pan and zoom
    ax = fig.add_subplot(1, 1, 1)
    ax.plot((1, 2, 1))
    plt.show()

Known limitations:
    - Only support linear and log scale axes.
    - Zoom area not working well with keep aspect ratio.
    - Interfere with matplotlib toolbar.
"""

from typing import Any

import math
import weakref
import logging
import warnings

import matplotlib.pyplot as _plt
import numpy

from PySide6.QtCore import Qt
from PySide6.QtGui import QIcon, QAction
from PySide6.QtWidgets import QLabel

from ... import Dict

warnings.filterwarnings(
    "ignore",
    message=r".*Ignoring fixed .* limits to fulfill fixed data aspect.*",
)

__all__ = ['figure_pz', 'MplInteraction', 'PanAndZoom']


class MplInteraction(object):
    """Base class for class providing interaction to a matplotlib Figure."""

    def __init__(self, figure):
        """
        Args:
            figure (figure): The matplotlib figure to attach the behavior to.
        """
        self._fig_ref = weakref.ref(figure)
        self._cids = []

    def __del__(self):
        """Disconnect."""
        self.disconnect()

    def _add_connection(self, event_name, callback):
        """Called to add a connection to an event of the figure.

        Args:
            event_name (str): The matplotlib event name to connect to.
            callback (callback): The callback to register to this event.
        """
        cid = self.figure.canvas.mpl_connect(event_name, callback)
        self._cids.append(cid)

    def disconnect(self):
        """Disconnect interaction from Figure."""
        if self._fig_ref is not None:
            figure = self._fig_ref()
            if figure is not None:
                for cid in self._cids:
                    figure.canvas.mpl_disconnect(cid)
            self._fig_ref = None

    @property
    def figure(self):
        """The Figure this interaction is connected to or None if not
        connected."""
        return self._fig_ref() if self._fig_ref is not None else None

    def _axes_to_update(self, event):
        """Returns two sets of Axes to update according to event. Takes care of
        multiple axes and shared axes.

        Args:
            event (MouseEvent): Matplotlib event to consider

        Returns:
            tuple: Axes for which to update xlimits and ylimits.
            2-tuple of set (xaxes, yaxes)
        """
        x_axes, y_axes = set(), set()

        # Go through all axes to enable zoom for multiple axes subplots
        for ax in self.figure.axes:
            if ax.contains(event)[0]:
                # For twin x axes, makes sure the zoom is applied once
                shared_x_axes = set(ax.get_shared_x_axes().get_siblings(ax))
                if x_axes.isdisjoint(shared_x_axes):
                    x_axes.add(ax)

                # For twin y axes, makes sure the zoom is applied once
                shared_y_axes = set(ax.get_shared_y_axes().get_siblings(ax))
                if y_axes.isdisjoint(shared_y_axes):
                    y_axes.add(ax)

        return x_axes, y_axes

    def _draw(self):
        """Convenient method to redraw the figure."""
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=
                r".*Ignoring fixed .* limits to fulfill fixed data aspect.*",
            )
            self.figure.canvas.draw()


class ZoomOnWheel(MplInteraction):
    """Class providing zoom on wheel interaction to a matplotlib Figure.

    This class extends the `MplInteraction` class.

    Supports subplots, twin Axes and log scales.
    """

    def __init__(self, figure=None, scale_factor=1.1):
        """
        Args:
            figure (figure): The matplotlib figure to attach the behavior to.
            scale_factor (float): The scale factor to apply on wheel event.
        """
        super(ZoomOnWheel, self).__init__(figure)
        self._add_connection('scroll_event', self._on_mouse_wheel)

        self.scale_factor = scale_factor

    @staticmethod
    def _zoom_range(begin, end, center, scale_factor, scale):
        """Compute a 1D range zoomed around center.

        Args:
            begin (float): The begin bound of the range
            end (float): The end bound of the range
            center (float): The center of the zoom (i.e., invariant point)
            scale_factor (float): The scale factor to apply
            scale (str): The scale of the axis

        Returns:
            tuple: The zoomed range (min, max)
        """
        if begin < end:
            min_, max_ = begin, end
        else:
            min_, max_ = end, begin

        if scale == 'linear':
            old_min, old_max = min_, max_
        elif scale == 'log':
            old_min = numpy.log10(min_ if min_ > 0. else numpy.nextafter(0, 1))
            center = numpy.log10(center if center >
                                 0. else numpy.nextafter(0, 1))
            old_max = numpy.log10(max_) if max_ > 0. else 0.
        else:
            logging.warning('Zoom on wheel not implemented for scale "%s"' %
                            scale)
            return begin, end

        offset = (center - old_min) / (old_max - old_min)
        range_ = (old_max - old_min) / scale_factor
        new_min = center - offset * range_
        new_max = center + (1. - offset) * range_

        if scale == 'log':
            try:
                new_min, new_max = 10.**float(new_min), 10.**float(new_max)
            except OverflowError:  # Limit case
                new_min, new_max = min_, max_
            if new_min <= 0. or new_max <= 0.:  # Limit case
                new_min, new_max = min_, max_

        if begin < end:
            return new_min, new_max
        else:
            return new_max, new_min

    def _on_mouse_wheel(self, event: Any) -> None:
        """Handle mouse wheel zoom for all relevant axes.

        The zoom is centered on the mouse cursor (in data coordinates) and applied
        to all x and y axes returned by :meth:`_axes_to_update`.

        Args:
            event: Backend-specific mouse wheel event. Expected to provide:
                - ``x`` and ``y``: mouse position in display (pixel) coordinates.
                - ``step``: signed scroll step (positive for zoom in, negative for
                zoom out).
        """
        if not hasattr(event, "step") or event.step == 0:
            return

        scale_factor: float = (self.scale_factor if event.step > 0 else 1.0 /
                               self.scale_factor)

        # Determine which axes to zoom.
        x_axes, y_axes = self._axes_to_update(event)

        # Nothing to update.
        if not x_axes and not y_axes:
            return

        # Use the first available axis as a reference to convert display
        # coordinates (pixels) to data coordinates.
        axes_for_reference = x_axes or y_axes
        reference_ax = next(iter(axes_for_reference))

        transform = reference_ax.transData.inverted()
        xdata, ydata = transform.transform_point((event.x, event.y))

        def _zoom_axis_limits(ax: Any, is_x: bool, data_coord: float) -> None:
            """Zoom a single axis while suppressing aspect-ratio warnings.

            Args:
                ax: Matplotlib axes object.
                is_x: If True, zoom the x-axis; otherwise zoom the y-axis.
                data_coord: Data coordinate of the mouse position along this axis.
            """
            if is_x:
                current_min, current_max = ax.get_xlim()
                scale = ax.get_xscale()
                set_limits = ax.set_xlim
            else:
                current_min, current_max = ax.get_ylim()
                scale = ax.get_yscale()
                set_limits = ax.set_ylim

            new_min, new_max = self._zoom_range(
                current_min,
                current_max,
                data_coord,
                scale_factor,
                scale,
            )

            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    message=
                    (".*Ignoring fixed .* limits to fulfill fixed data aspect.*"
                    ),
                )
                set_limits((new_min, new_max))

        # Apply zoom to all x-axes.
        for ax in x_axes:
            _zoom_axis_limits(ax, is_x=True, data_coord=xdata)

        # Apply zoom to all y-axes.
        for ax in y_axes:
            _zoom_axis_limits(ax, is_x=False, data_coord=ydata)

        # Redraw once after all axes are updated.
        self._draw()


class PanAndZoom(ZoomOnWheel):
    """Class providing pan & zoom interaction to a matplotlib Figure. Left
    button for pan, right button for zoom area and zoom on wheel. Support
    subplots, twin Axes and log scales.

    This class extends the `ZoomOnWheel` class.
    """

    def __init__(self, figure=None, scale_factor=1.1):
        """
        Args:
            figure (figure): The matplotlib figure to attach the behavior to.
            scale_factor (float): The scale factor to apply on wheel event.
        """
        super(PanAndZoom, self).__init__(figure, scale_factor)
        self._add_connection('button_press_event', self._on_mouse_press)
        self._add_connection('button_release_event', self._on_mouse_release)
        self._add_connection('motion_notify_event', self._on_mouse_motion)

        self._pressed_button = None  # To store active button
        self._axes = None  # To store x and y axes concerned by interaction
        self._event = None  # To store reference event during interaction

        self.options = Dict(dict(report_point_position=True,))
        self.logger = None
        self._statusbar_label = None

        # self._get_images_path()
        # self._add_toolbar_tools()
        self._style_figure()

        self._ix_iy_old = (0, 0)

    def _get_images_path(self):
        """Get the path to images.

        Returns:
            str: path

        Raises:
            Exception: path error
        """
        # to be removed
        try:  # Get tool image path
            from pathlib import Path
            from ... import _gui
            imgs_path = Path(_gui.__file__).parent / '_imgs'
            if imgs_path.is_dir() == False:
                print(f'Bad File path for images! {imgs_path}')
                imgs_path = None
        except Exception as e:
            print('ERROR: ', e)
            imgs_path = None
        self.imgs_path = imgs_path

        return imgs_path

    def _add_toolbar_tools(self):
        """Add tools."""
        # TODO: Outdated - to be removed

        from matplotlib.backend_tools import ToolToggleBase  # ToolBase

        class ToolPointPosition(ToolToggleBase):
            """Tools."""
            default_keymap = 'Ctrl+p'
            description = 'Click to get point coordinate printed'
            default_toggled = False
            image = None  # str(imgs_path)

            def __init__(self, *args, parent=None, **kwargs):
                super().__init__(*args, **kwargs)
                if parent is None:
                    raise ('Pass a parent')
                self.parent = parent

            def enable(self, *args):
                self.parent.options.report_point_position = True

            def disable(self, *args):
                self.parent.options.report_point_position = False

        fig = self.figure
        imgs_path = self.imgs_path
        # pylint: disable=attribute-defined-outside-init
        toolbar = self.toolbar = fig.canvas.manager.toolbar

        # Get tool manager

        # TODO: Remove use of tool manager just use PySide2 bare as below

        # ToolbarQt   --- https://github.com/matplotlib/matplotlib/blob/master/lib/matplotlib/backends/backend_qt5.py
        tm = fig.canvas.manager.toolmanager

        self.tm = tm
        # Tool: Print point location
        ToolPointPosition.image = str(imgs_path / 'click.png')
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            tm.add_tool("Point_position", ToolPointPosition, parent=self)
        fig.canvas.manager.toolbar.add_tool(tm.get_tool("Point_position"),
                                            "toolgroup")

        # Tool: Copy to Clipboard
        from matplotlib.backend_tools import ToolCopyToClipboard
        ToolCopyToClipboard.image = str(imgs_path / 'copy.png')
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # OVvrwties Ctrl+C and issues warning
            tm.add_tool("Copy_to_clipboard", ToolCopyToClipboard)
        fig.canvas.manager.toolbar.add_tool(tm.get_tool("Copy_to_clipboard"),
                                            "toolgroup")

        if 1:  # add QT Pieces
            toolbar.action_ascale = QAction(
                QIcon(str(imgs_path / 'auto_zoom.png')), 'Auto scale', toolbar)
            toolbar.action_ascale.setShortcut('A')
            toolbar.action_ascale.setShortcutContext(Qt.WindowShortcut)
            toolbar.action_ascale.setStatusTip('Autoscale')
            toolbar.action_ascale.triggered.connect(self.auto_scale)
            toolbar.addAction(toolbar.action_ascale)

        # Status Bar: Second label to report
        figManager = fig.canvas.manager  # plt.get_current_fig_manager()
        status_bar = figManager.window.statusBar()
        self._status_label_2 = QLabel(status_bar)
        self._status_label_2.setText('')
        status_bar.addWidget(self._status_label_2)
        #from matplotlib.backends.backend_qt5 import StatusbarQt
        #st = StatusbarQt(figManager.window, figManager.toolmanager)
        # figManager.statusbar.set_message('')

    def auto_scale(self):
        """Auto scaler."""
        for ax in self.figure.axes:
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    message=
                    r".*Ignoring fixed .* limits to fulfill fixed data aspect.*",
                )
                ax.autoscale()
        # self.figure.canvas.flush_events()
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=
                r".*Ignoring fixed .* limits to fulfill fixed data aspect.*",
            )
            self.figure.canvas.draw()

    def _style_figure(self):
        """Style figure."""
        #self.figure.dpi = 150
        pass

    @staticmethod
    def _pan_update_limits(ax, axis_id, event, last_event):
        """Compute limits with applied pan.

        Args:
            axis_id (int): ID of the axis
            event (event): The event
            last_event (event): The previous event

        Returns:
            double: New limit

        Raises:
            ValueError: Value error
            OverflowError: Overflow error
        """
        assert axis_id in (0, 1)
        if axis_id == 0:
            lim = ax.get_xlim()
            scale = ax.get_xscale()
        else:
            lim = ax.get_ylim()
            scale = ax.get_yscale()

        pixel_to_data = ax.transData.inverted()
        data = pixel_to_data.transform_point((event.x, event.y))
        last_data = pixel_to_data.transform_point((last_event.x, last_event.y))

        if scale == 'linear':
            delta = data[axis_id] - last_data[axis_id]
            new_lim = lim[0] - delta, lim[1] - delta
        elif scale == 'log':
            try:
                delta = math.log10(data[axis_id]) - \
                    math.log10(last_data[axis_id])
                new_lim = [
                    pow(10., (math.log10(lim[0]) - delta)),
                    pow(10., (math.log10(lim[1]) - delta))
                ]
            except (ValueError, OverflowError):
                new_lim = lim  # Keep previous limits
        else:
            logging.warning('Pan not implemented for scale "%s"' % scale)
            new_lim = lim
        return new_lim

    def _pan(self, event):
        """Pan.

        Args:
            event (event): The event
        """
        if event.name == 'button_press_event':  # begin pan
            self._event = event

        elif event.name == 'button_release_event':  # end pan
            self._event = None

        elif event.name == 'motion_notify_event':  # pan
            if self._event is None:
                return

            if event.x != self._event.x:
                for ax in self._axes[0]:
                    xlim = self._pan_update_limits(ax, 0, event, self._event)
                    ax.set_xlim(xlim)

            if event.y != self._event.y:
                for ax in self._axes[1]:
                    ylim = self._pan_update_limits(ax, 1, event, self._event)
                    ax.set_ylim(ylim)

            if event.x != self._event.x or event.y != self._event.y:
                self._draw()

            self._event = event

    def _zoom_area(self, event):
        """Zoom.

        Args:
            event (event): The event
        """
        if event.name == 'button_press_event':  # begin drag
            self._event = event
            # pylint: disable=attribute-defined-outside-init
            self._patch = _plt.Rectangle(xy=(event.xdata, event.ydata),
                                         width=0,
                                         height=0,
                                         fill=False,
                                         linewidth=1.,
                                         linestyle='solid',
                                         color='black')
            self._event.inaxes.add_patch(self._patch)

        elif event.name == 'button_release_event':  # end drag
            self._patch.remove()
            del self._patch

            if (abs(event.x - self._event.x) < 3 or
                    abs(event.y - self._event.y) < 3):
                return  # No zoom when points are too close

            x_axes, y_axes = self._axes

            for ax in x_axes:
                pixel_to_data = ax.transData.inverted()
                begin_pt = pixel_to_data.transform_point((event.x, event.y))
                end_pt = pixel_to_data.transform_point(
                    (self._event.x, self._event.y))

                min_ = min(begin_pt[0], end_pt[0])
                max_ = max(begin_pt[0], end_pt[0])
                if not ax.xaxis_inverted():
                    ax.set_xlim(min_, max_)
                else:
                    ax.set_xlim(max_, min_)

            for ax in y_axes:
                pixel_to_data = ax.transData.inverted()
                begin_pt = pixel_to_data.transform_point((event.x, event.y))
                end_pt = pixel_to_data.transform_point(
                    (self._event.x, self._event.y))

                min_ = min(begin_pt[1], end_pt[1])
                max_ = max(begin_pt[1], end_pt[1])
                if not ax.yaxis_inverted():
                    ax.set_ylim(min_, max_)
                else:
                    ax.set_ylim(max_, min_)

            self._event = None

        elif event.name == 'motion_notify_event':  # drag
            if self._event is None:
                return

            if event.inaxes != self._event.inaxes:
                return  # Ignore event outside plot

            self._patch.set_width(event.xdata - self._event.xdata)
            self._patch.set_height(event.ydata - self._event.ydata)

        self._draw()

    def _on_mouse_press(self, event):
        """Mouse press event.

        Args:
            event (event): The event
        """
        if self._pressed_button is not None:
            return  # Discard event if a button is already pressed

        if event.button in (1, 3):  # Start
            x_axes, y_axes = self._axes_to_update(event)
            if x_axes or y_axes:
                self._axes = x_axes, y_axes
                self._pressed_button = event.button

                if self._pressed_button == 1:  # pan
                    self._pan(event)

                    if self.options.report_point_position:  # check if we want to report point
                        self._report_point_position(event)

                elif self._pressed_button == 3:  # zoom area
                    self._zoom_area(event)

    def _on_mouse_release(self, event):
        """Mouse release event.

        Args:
            event (event): The event
        """
        if self._pressed_button == event.button:
            if self._pressed_button == 1:  # pan
                self._pan(event)
            elif self._pressed_button == 3:  # zoom area
                self._zoom_area(event)
            self._pressed_button = None

    def _on_mouse_motion(self, event):
        """Mouse motion event.

        Args:
            event (event): The event
        """
        if self._pressed_button == 1:  # pan
            self._pan(event)
        elif self._pressed_button == 3:  # zoom area
            self._zoom_area(event)

    def _report_point_position(self, event):
        """Report point position.

        Args:
            event (event): the event
        """
        ix, iy = event.xdata, event.ydata

        if hasattr(self, '_ix_iy_old'):
            ix_old, iy_old = self._ix_iy_old
        else:
            ix_old, iy_old = (ix, iy)

        self._ix_iy_old = ix, iy

        _text = f'(x,y) = ({ix:.4f}, {iy:.4f})  Δ last point ({ix-ix_old:.4f}, {iy-iy_old:.4f})'
        if self.logger:
            self.logger.info(_text)
        if self._statusbar_label:
            self._statusbar_label.setText(_text)
        # print(_text)


def figure_pz(*args, **kwargs):
    """matplotlib.pyplot.figure with pan and zoom interaction."""
    #import warnings
    # warnings.filterwarnings(action='ignore')

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        fig = _plt.figure(*args, **kwargs)
    fig.pan_zoom = PanAndZoom(fig)

    # warnings.resetwarnings()

    return fig


# """
# if __name__ == "__main__":
#     import matplotlib.pyplot as plt

#     fig = figure_pz()
#     # Alternative:
#     # fig = plt.figure()
#     # pan_zoom = PanAndZoom(fig)

#     nrow, ncol = 2, 3

#     ax1 = fig.add_subplot(nrow, ncol, 1)
#     ax1.set_title('basic')
#     ax1.plot((1, 2, 3))

#     ax2 = fig.add_subplot(nrow, ncol, 2)
#     ax2.set_title('log + twinx')
#     ax2.set_yscale('log')
#     ax2.plot((1, 2, 1))

#     ax2bis = ax2.twinx()
#     ax2bis.plot((3, 2, 1), color='red')

#     ax3 = fig.add_subplot(nrow, ncol, 3)
#     ax3.set_title('inverted y axis')
#     ax3.plot((1, 2, 3))
#     lim = ax3.get_ylim()
#     ax3.set_ylim(lim[1], lim[0])

#     ax4 = fig.add_subplot(nrow, ncol, 4)
#     ax4.set_title('keep ratio')
#     ax4.axis('equal')
#     ax4.imshow(numpy.arange(100).reshape(10, 10))

#     ax5 = fig.add_subplot(nrow, ncol, 5)
#     ax5.set_xlabel('symlog scale + twiny')
#     ax5.set_xscale('symlog')
#     ax5.plot((1, 2, 3))
#     ax5bis = ax5.twiny()
#     ax5bis.plot((3, 2, 1), color='red')

#     # The following is taken from:
#     # http://matplotlib.org/examples/axes_grid/demo_curvelinear_grid.html
#     from mpl_toolkits.axisartist import Subplot
#     from mpl_toolkits.axisartist.grid_helper_curvelinear import \
#         GridHelperCurveLinear

#     def tr(x, y):  # source (data) to target (rectilinear plot) coordinates
#         x, y = numpy.asarray(x), numpy.asarray(y)
#         return x + 0.2 * y, y - x

#     def inv_tr(x, y):
#         x, y = numpy.asarray(x), numpy.asarray(y)
#         return x - 0.2 * y, y + x

#     grid_helper = GridHelperCurveLinear((tr, inv_tr))

#     ax6 = Subplot(fig, nrow, ncol, 6, grid_helper=grid_helper)
#     fig.add_subplot(ax6)
#     ax6.set_title('non-ortho axes')

#     xx, yy = tr([3, 6], [5.0, 10.])
#     ax6.plot(xx, yy)

#     ax6.set_aspect(1.)
#     ax6.set_xlim(0, 10.)
#     ax6.set_ylim(0, 10.)

#     ax6.axis["t"] = ax6.new_floating_axis(0, 3.)
#     ax6.axis["t2"] = ax6.new_floating_axis(1, 7.)
#     ax6.grid(True)

#     plt.show()
# """



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import numpy as np
from qiskit_metal import Dict
from qiskit_metal.qlibrary.core import QRoutePoint

from collections import OrderedDict
from .meandered import RouteMeander
from .pathfinder import RoutePathfinder


# class RouteMixed(RouteFramed, RoutePathfinder, RouteMeander):
class RouteMixed(RoutePathfinder, RouteMeander):
    """Implements fully featured Routing, allowing different type of
    connections between anchors. The comprehensive Routing class. Inherits
    `RoutePathfinder, RouteMeander` class, thus also QRoute and RouteAnchors.

    .. meta::
        :description: Route Mixed

    Default Options:
        * between_anchors: Empty OrderedDict -- Intermediate anchors only; doesn't include endpoints

    QRoute Default Options:
        * pin_inputs: Dict
            * start_pin: Dict -- Component and pin string pair. Define which pin to start from
                * component: '' -- Name of component to start from, which has a pin
                * pin: '' -- Name of pin used for pin_start
            * end_pin=Dict -- Component and pin string pair. Define which pin to start from
                * component: '' -- Name of component to end on, which has a pin
                * pin: '' -- Name of pin used for pin_end
        * lead: Dict
            * start_straight: '0mm' -- Lead-in, defined as the straight segment extension from start_pin.  Defaults to 0.1um.
            * end_straight: '0mm' -- Lead-out, defined as the straight segment extension from end_pin.  Defaults to 0.1um.
            * start_jogged_extension: '' -- Lead-in, jogged extension of lead-in. Described as list of tuples
            * end_jogged_extension: '' -- Lead-out, jogged extension of lead-out. Described as list of tuples
        * fillet: '0'
        * total_length: '7mm'
        * trace_width: 'cpw_width' -- Defines the width of the line

    RouteAnchors Default Options:
        * anchors: OrderedDict -- Intermediate anchors only; doesn't include endpoints

    RoutePathfinder Default Options:
        * step_size: '0.25mm' -- Length of the step for the A* pathfinding algorithm
        * advanced: Dict
            * avoid_collision: 'true' -- true/false, defines if the route needs to avoid collisions.  Defaults to 'true'.

    RouteMeander Default Options:
        * meander: Dict
            * spacing: '200um' -- Minimum spacing between adjacent meander curves
            * asymmetry='0um' -- Offset between the center-line of the meander and the center-line that stretches from the tip of lead-in to the x (or y) coordinate of the tip of the lead-out.  Defaults to '0um'.
        * snap: 'true'
        * prevent_short_edges: 'true'

    How to specify between_anchors for the RouteMixed
        `between_anchors` have to be specified in an OrderedDict with incremental keys.
        the value of each key specifies the type of routing algorythm to run between each anchors
        and between the pins and the first and last anchors.

        For example, for a RouteMixed with 4 anchors:

        .. code-block:: python
            :linenos:

            between_anchors = OrderedDict()
            between_anchors[0] = "S"
            between_anchors[1] = "M"
            between_anchors[2] = "S"
            between_anchors[3] = "M"
            between_anchors[4] = "PF"

        * "S" = Utilizes the RouteStraight methods
        * "M" = Utilizes the RouteMeander methods
        * "PF" = Utilizes the RoutePathfinder methods
    """

    default_options = Dict(
        between_anchors=OrderedDict(
        ),  # Intermediate anchors only; doesn't include endpoints
        # Example: {1: "M", 2: "S", 3: "PF"}
        # startpin -> startpin + leadin -> anchors -> endpin + leadout -> endpin
    )
    """Default options"""

    TOOLTIP = """Implements fully featured Routing, allowing different type of
    connections between anchors."""

    def make(self):
        """Generates path from start pin to end pin."""
        p = self.parse_options()
        anchors = p.anchors
        between_anchors = p.between_anchors

        # Set the CPW pins and add the points/directions to the lead-in/out arrays
        self.set_pin("start")
        self.set_pin("end")

        # Align the lead-in/out to the input options set from the user
        start_point = self.set_lead("start")
        end_point = self.set_lead("end")

        # approximate length needed for individual meanders
        # the meander algorithm directly reads from self._length_segment
        count_meanders_list = [
            1 if x == "M" else 0 for x in list(between_anchors.values())
        ]
        self._length_segment = None
        if any(count_meanders_list):
            self._length_segment = ((self.p.total_length - (self.head.length + self.tail.length) \
                                   - self.free_manhattan_length_anchors()) / sum(count_meanders_list)) \
                                   + (self.free_manhattan_length_anchors() / len(count_meanders_list))

        # find the points to connect between each pair of anchors, or between anchors and leads
        # at first, store points "per segment" in a dictionary, so it is easier to apply length requirements
        self.intermediate_pts = OrderedDict()
        meanders = set()
        for arc_num, coord in anchors.items():
            # determine what is the connection strategy for this pair, based on user inputs
            connect_method = self.select_connect_method(arc_num)
            if connect_method == self.connect_meandered:
                meanders.add(arc_num)
            # compute points connecting the anchors, all but the last
            arc_pts = connect_method(self.get_tip(), QRoutePoint(coord))
            if arc_pts is None:
                self.intermediate_pts[arc_num] = [coord]
            else:
                self.intermediate_pts[arc_num] = np.concatenate(
                    [arc_pts, [coord]], axis=0)
        # compute last connection point to the output QRouteLead
        connect_method = self.select_connect_method(len(anchors))
        if connect_method == self.connect_meandered:
            meanders.add(len(anchors))
        arc_pts = connect_method(self.get_tip(), end_point)
        if arc_pts is not None:
            self.intermediate_pts[len(anchors)] = np.array(arc_pts)

        # concatenate all points, transforming the dictionary into a single numpy array
        self.trim_pts()
        dictionary_intermediate_pts = self.intermediate_pts
        self.intermediate_pts = np.concatenate(list(
            self.intermediate_pts.values()),
                                               axis=0)

        if any(count_meanders_list):
            # refine length of meanders
            total_delta_length = self.p.total_length - self.length
            individual_delta_length = total_delta_length / len(meanders)
            for m in meanders:
                arc_pts = dictionary_intermediate_pts[m][:-1]
                if m == 0:
                    meander_start_point = start_point
                else:
                    meander_start_point = QRoutePoint(anchors[m - 1])
                if m == len(anchors):
                    meander_end_point = end_point
                else:
                    meander_end_point = QRoutePoint(anchors[m])
                dictionary_intermediate_pts[m] = self.adjust_length(
                    individual_delta_length, arc_pts, meander_start_point,
                    meander_end_point)
                dictionary_intermediate_pts[m] = np.concatenate(
                    [dictionary_intermediate_pts[m], [anchors[m]]], axis=0)
        self.intermediate_pts = np.concatenate(list(
            dictionary_intermediate_pts.values()),
                                               axis=0)

        # Make points into elements
        self.make_elements(self.get_points())

    def select_connect_method(self, segment_num):
        """Translates the user-selected connection method into the right method
        to execute.

        Args:
            segment_num (int): Segment ID. Counts 0 as the first segment after the lead-in.

        Return:
            object: selected method
        """
        between_anchors = self.parse_options().between_anchors
        if segment_num in between_anchors:
            type_connect = between_anchors[segment_num]
        else:
            type_connect = "S"
        # print(type_connect)
        if type_connect == "S":
            return self.connect_simple
        if type_connect == "PF":
            return self.connect_astar_or_simple
        if type_connect == "M":
            return self.connect_meandered



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

#pylint: disable-msg=unnecessary-pass
#pylint: disable-msg=protected-access
#pylint: disable-msg=broad-except
"""Qiskit Metal unit tests analyses functionality."""

import unittest
from qiskit_metal.toolbox_python.display import Headings
from qiskit_metal.toolbox_python.display import Color
from qiskit_metal.toolbox_python.display import MetalTutorialMagics
from qiskit_metal.toolbox_python import display
from qiskit_metal.toolbox_python import utility_functions
from qiskit_metal.toolbox_python._logging import LogStore


class TestToolboxPython(unittest.TestCase):
    """Unit test class."""

    def setUp(self):
        """Setup unit test."""
        pass

    def tearDown(self):
        """Tie any loose ends."""
        pass

    def test_instantiate_log_store(self):
        """Test instantiation of LogStore class."""
        try:
            LogStore('my title', 100)
        except Exception:
            self.fail("LogStore failed")

    def test_instantiate_headings(self):
        """Test instantiation of Headings class."""
        try:
            Headings()
        except Exception:
            self.fail("Headings failed")

    def test_instantiate_metal_tutorial_magics(self):
        """Test instantiation of MetalTutorialMagics class."""
        try:
            MetalTutorialMagics()
        except Exception:
            self.fail("MetalTutorialMagics failed")

    def test_instantiate_color(self):
        """Test instantiation of Color class."""
        try:
            Color()
        except Exception:
            self.fail("Color failed")

    def test_display_style_colon_list(self):
        """Test style_colon_list in display.py."""
        actual = display.style_colon_list("text:more:still")
        expected = "text:\033[94mmore:still\033[0m"

        self.assertEqual(actual, expected)

    def test_display_format_dict_ala_z(self):
        """Test functionality of format_dict_ala_z in display.py."""
        my_dict = {'first': 1}
        expected = "  'first'             : 1,                            "
        actual = display.format_dict_ala_z(my_dict)

        self.assertEqual(expected, actual)

    def test_utility_defaults(self):
        """Test defaults in utility_functions.py."""
        self.assertEqual(utility_functions._old_warn, None)

    def test_utility_dict_start_with(self):
        """Test dict_start_with in utility_functions.py."""
        my_dict = {
            'first': 1,
            'second': 2,
            'third': {
                'third-one': '3-1',
                'third-two': {
                    'third-two-only': 'uh'
                }
            },
            'fourth': 4
        }

        expected = [[{
            'third-one': '3-1',
            'third-two': {
                'third-two-only': 'uh'
            }
        }], {
            'third': {
                'third-one': '3-1',
                'third-two': {
                    'third-two-only': 'uh'
                }
            }
        }, [], {}]

        actual = []
        actual.append(utility_functions.dict_start_with(my_dict, 'third'))
        actual.append(
            utility_functions.dict_start_with(my_dict, 'third', as_=dict))
        actual.append(utility_functions.dict_start_with(my_dict, 'nope'))
        actual.append(
            utility_functions.dict_start_with(my_dict, 'nope', as_=dict))

        for i in range(4):
            self.assertEqual(actual[i], expected[i])

    def test_utility_data_frame_empty_typed(self):
        """Test data_frame_empty_typed in utility_functions.py."""
        base = dict(col1=bool, col2=object)

        df = utility_functions.data_frame_empty_typed(base)

        self.assertEqual(df.dtypes['col1'], bool)
        self.assertEqual(df.dtypes['col2'], object)

    def test_utility_copy_update(self):
        """Test functionality of copy_update in utility_functions.py."""
        self.assertEqual(utility_functions.copy_update({'a': 1}, {}), {'a': 1})
        self.assertEqual(utility_functions.copy_update({'a': 1}, {'b': 2}), {
            'a': 1,
            'b': 2
        })
        self.assertEqual(utility_functions.copy_update({'a': 1}, {'a': 2}),
                         {'a': 2})

    def test_utility_bad_fillet_idxs(self):
        """Test functionality of bad_fillet_idxs in utility_functions.py."""
        results = utility_functions.bad_fillet_idxs([(1.0, 1.0), (1.5, 1.5),
                                                     (1.51, 1.5), (2.0, 2.0)],
                                                    0.1)
        self.assertEqual(results, [1, 2])

    def test_utility_clean_name(self):
        """Test clean_name in utility_function.py."""
        self.assertEqual(
            utility_functions.clean_name('32v2 g #Gmw845h$W b53wi '),
            '_32v2_g__Gmw845h_W_b53wi_')
        self.assertEqual(utility_functions.clean_name('halestorm rulz!'),
                         'halestorm_rulz_')

    def test_utility_toggle_numbers(self):
        """Test functionality of toggle_numbers in utility_functions.py."""
        self.assertEqual(utility_functions.toggle_numbers([3, 5, 9, 1], 14),
                         [0, 1, 2, 4, 6, 7, 8, 11, 12, 13])

    def test_utility_get_range_of_vertex_to_not_fillet(self):
        """Test functionality of get_range_of_vertex_to_not_fillet in
        utility_functions.py."""
        my_list = [(1, 1), (1, 2), (1, 2), (2, 2), (5, 5), (3, 2), (11, 11),
                   (11, 11), (11, 21), (12, 21)]
        result = utility_functions.get_range_of_vertex_to_not_fillet(
            my_list, 0.25)
        self.assertEqual(result, [(0, 2), (6, 7)])

    def test_utility_compress_vertex_list(self):
        """Test functionality of compress_vertex_list in
        utility_functions.py."""
        my_list = [1, 2, 5, 2, 6, 7, 10, 4, 7, 1]
        expected = [(1, 1), (1, 2), (2, 2), (4, 7), (7, 7), (10, 10)]
        self.assertEqual(utility_functions.compress_vertex_list(my_list),
                         expected)


if __name__ == '__main__':
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Main module that handles the entier plot window which is docked inside the
main window.

This can be undocked and can have its own toolbar. this is largley why I
decided to use a QMainWindow, so that we can have inner docking and
toolbars available.
"""

from typing import TYPE_CHECKING

from PySide6 import QtWidgets
from PySide6.QtWidgets import (QApplication, QFileDialog, QLabel, QMainWindow,
                               QMessageBox)

from ... import config
if not config.is_building_docs():
    # Only import PlotCanvas if the docs are NOT being built
    from ....renderers.renderer_mpl.mpl_canvas import PlotCanvas

from ...plot_window_ui import Ui_MainWindowPlot

if TYPE_CHECKING:
    # https://stackoverflow.com/questions/39740632/python-type-hinting-without-cyclic-imports
    from ...main_window import MetalGUI, QMainWindowExtension


class QMainWindowPlot(QMainWindow):
    """This is just a handler (container) for the UI; it a child object of the
    main gui.

    Extends the `QMainWindow` class

    PySide2 Signal / Slots Extensions:
        The UI can call up to this class to execeute button clicks for instance
        Extensiosn in qt designer on signals/slots are linked to this class

    Core canvas plot widget:
        canvas: The core plot object. Can be mpl or any other renderer.
    """

    def __init__(self, gui: 'MetalGUI', parent_window: 'QMainWindowExtension'):
        """
        Args:
            gui (MetalGUI): The GUI
            parent_window (QMainWindowExtension): Parent window
        """
        # Q Main WIndow
        super().__init__(parent_window)

        # Parent GUI related
        self.gui = gui
        self.logger = gui.logger
        self.statusbar_label = gui.statusbar_label

        # UI
        self.ui = Ui_MainWindowPlot()
        self.ui.setupUi(self)

        self.statusBar().hide()

        # Add MPL plot widget to window
        # Core object -- the center of this entire widget
        self.canvas = PlotCanvas(self.design,
                                 self,
                                 logger=self.logger,
                                 statusbar_label=self.statusbar_label)

        self.ui.centralwidget.layout().addWidget(self.canvas)

    def set_design(self, design):
        """Set the design.

        Args:
            design (QDesign): Design to set the canvas to
        """
        self.canvas.set_design(design)

    @property
    def design(self):
        """Returns the design."""
        return self.gui.design

    def replot(self):
        """Tells the canvas to replot."""
        # self.logger.debug("Force replot")
        self.canvas.plot()

    def auto_scale(self):
        """Tells the canvas to perform an automatic scale."""
        self.logger.debug("Autoscale")
        self.canvas.auto_scale()

    def pan(self):
        """Displays a message about how to pan."""
        QMessageBox.about(
            self, "Pan", """Navigation help:

Pan:
(click and drag)
Click and drag the plot screen.

Zoom:
(scroll, or right click and drag)
Either use the mouse middle wheel to zoom in and out by scrolling,
or use the right click and drag to select a region.""")

    def zoom(self):
        """Displays a message about how to zoom."""
        QMessageBox.about(
            self, "Zoom", "Either use the mouse middle wheel"
            " to zoom in and out by scrolling, or use the right click and"
            " drag to select a region.")

    def set_position_track(self, yesno: bool):
        """Set the position tracker.

        Args:
            yesno (bool): Whether or not to display instructions
        """
        if yesno:
            self.logger.info("Click a point in the plot window to see"
                             " its coordinate.")
        self.canvas.panzoom.options.report_point_position = yesno

    def set_show_pins(self, yesno: bool):
        """Displays on the logger whether or not pins are showing.

        Args:
            yesno (bool): Whether or not to show pins
        """
        self.logger.info(f"Showing pins: {yesno}")



from qiskit_metal import Dict
from typing import Tuple, Union


class Sweeper():
    """The methods allow users to sweep a variable in a components's options.
    Need access to renderers which are registered in QDesign."""

    def __init__(self, parent):
        """Connect to the QAnalysis child we sweeping can happen.

        Args:
            parent (QAnalysis): Will not be just QAnalysis, but a child or grandchild, etc. of QAnalysis.
        """

        # Reference to the instance (or child or grandchild) of QAnalysis.
        self.parent = parent

        #For easy access, make a reference to QDesign.
        if hasattr(self.parent, 'sim'):
            self.design = parent.sim.design
        elif hasattr(self.parent, 'design'):
            self.design = parent.design
        else:
            self.design = None

    def run_sweep(self, *args, **kwarg) -> Tuple[Dict, int]:
        """Ansys will be opened, if not already open, with an inserted project.  
        A design will be inserted by this method. 

        There are two ways to pass arguments.  You can use the previous run or 
        use updated arguments.  With both scenarios, qcomp_name, option_name, and 
        option_sweep must be passed.

        For the previous run, the arguments are all but the three required.  

        Args:
            qcomp_name (str): A component that contains the option to be swept.
            option_name (str): The option within qcomp_name to sweep.
            option_sweep (list): Each entry in the list is a value for
                        option_name.
            qcomp_render (list): The component to render to simulation software.
            open_terminations (list): Identify which kind of pins. Follow the
                        details from renderer QQ3DRenderer.render_design, or
                        QHFSSRenderer.render_design.
            port_list (list): List of tuples of jj's that shouldn't
                            be rendered.  Follow details from
                            renderer in QHFSSRenderer.render_design.
            jj_to_port (list): List of junctions (qcomp, qgeometry_name,
                                impedance, draw_ind) to render as lumped ports
                                or as lumped port in parallel with a sheet
                                inductance.    Follow details from renderer
                                in QHFSSRenderer.render_design.
            ignored_jjs (Union[list,None]): This is not used by all renderers,
                         just hfss.
            design_name(str): Name of design (workspace) to use in project.
            box_plus_buffer(bool): Render the entire chip or create a
                        box_plus_buffer around the components which are rendered.

        Returns:
            Tuple[Dict, int]: The dict key is each value of option_sweep, the
            value is the solution-data for each sweep.
            The int is the observation of searching for data from arguments as
            defined below.

            * 0 Have list of capacitance matrix.
            * 1 qcomp_name not registered in design.
            * 2 option_name is empty.
            * 3 option_name is not found as key in Dict.
            * 4 option_sweep is empty, need at least one entry.
            * 5 last key in option_name is not in Dict. 
            * 6 need to have at least three arguments
        """
        #Dict of all swept information.
        all_sweep = Dict()
        previous_run = Dict()
        clean_kwargs = Dict()
        use_previous_run = False

        # Decide if we use the previous run based on inputs given.
        if len(args) == 3 and len(kwarg) == 0:
            use_previous_run = True
            # Do not populate the previous_run if more than minimum required is passed.
            if hasattr(self.parent, 'sim'):
                if self.parent.sim.setup.run:
                    previous_run = self.parent.sim.setup.run
            elif self.parent.setup.run:
                previous_run = self.parent.setup.run

        if len(args) >= 3:
            option_path, a_value, check_result = self.error_check_sweep_input(
                args[0], args[1], args[2])
        else:
            return all_sweep, 6

        if check_result != 0:
            return all_sweep, check_result

        if len(args) > 3:
            clean_kwargs['components'] = args[3]
        if len(args) > 4:
            clean_kwargs['open_terminations'] = args[4]

        if 'design_name' in kwarg:
            #means use_previous_run=False
            clean_kwargs['name'] = kwarg['design_name']
            del kwarg['design_name']
        else:
            clean_kwargs['name'] = "Sweep_default"

        if use_previous_run and 'box_plus_buffer' not in previous_run:
            clean_kwargs['box_plus_buffer'] = True

        if not use_previous_run and 'box_plus_buffer' not in kwarg:
            clean_kwargs['box_plus_buffer'] = True

        all_dicts = {**previous_run, **clean_kwargs, **kwarg}

        all_sweep, check_result = self.iterate_option_sweep(
            args,
            all_dicts=all_dicts,
            option_path=option_path,
            a_value=a_value,
            all_sweep=all_sweep)

        return all_sweep, check_result

    def iterate_option_sweep(self, args: list, all_dicts: Dict,
                             option_path: list, a_value: Dict,
                             all_sweep: Dict) -> Tuple[Dict, int]:
        """Iterate through the values that user gave in option_sweep.  

        Args:
            args (Dict): Holds the three mandatory arguments in an expected sequence.
            all_dicts (Dict): User arguments manipulated to account for using previous_run. 
            option_path (list):  The list has traversed the option Dict.
            a_value (Dict): Has the value from the dictionary of the searched key.
            all_sweep (Dict): Will be populated during the iteration. 

        Returns:
            Tuple[Dict, int]: The dict key is each value of option_sweep, the
            value is the solution-data for each sweep.
            The int is the observation of searching for data from arguments as
            defined below.

                * 0 Have list of capacitance matrix.
                * 5 last key in option_name is not in Dict. 
        """

        for _, item in enumerate(args[2]):
            # Last item in list.
            if option_path[-1] in a_value.keys():
                a_value[option_path[-1]] = item
            else:
                self.design.logger.warning(
                    f'Key="{option_path[-1]}" is not in dict.')
                return all_sweep, 5

            self.design.rebuild()

            try:
                self.parent.run(**all_dicts)
            except Exception as ex:
                template = "An exception of type {0} occurred. Arguments:\n{1!r}"
                message = template.format(type(ex).__name__, ex.args)
                self.design.logger.warning(
                    f'For class {self.parent.__class__.__name__}, '
                    f'option_name={".".join(option_path)}, key={item}, '
                    f'run() did not execute as expected: {message}')

            self.populate_all_sweep(all_sweep, item, args[1])

        return all_sweep, 0

    # #######  Populate all_sweep
    def populate_all_sweep(self, all_sweep: Dict, item: str, option_name: str):
        """Populate the Dict passed in all_sweep from QAnalysis.  

        Args:
            all_sweep (Dict): Reference to hold each item which corresponds
              to option_name.
            item (str): The value of each item that we want to sweep
              in option_name.
            option_name (str): The option of QComponent that we want to sweep.
        """
        sweep_values = Dict()
        sweep_values['option_name'] = option_name
        sweep_values['variables'] = self.parent._variables

        if hasattr(self.parent, 'sim'):
            sweep_values['sim_variables'] = self.parent.sim._variables

        all_sweep[item] = sweep_values

    # ####### Error checking user input.

    def error_check_sweep_input(self, qcomp_name: str, option_name: str,
                                option_sweep: list) -> Tuple[list, Dict, int]:
        """ Implement error checking of data for sweeping.

        Args:
            qcomp_name (str): Component that contains the option to be swept.
            option_name (str): The option within qcomp_name to sweep.
            option_sweep (list): Each entry in the list is a value
                                for option_name.

        Returns:
            Tuple[list, Dict, int]:
            The list has traversed the option Dict.
            addict.addict.Dict has the value from the dictionary of the searched key.
            The int is the observation of searching for data from arguments.

            * 0 Error not detected in the input-data.
            * 1 qcomp_name not registered in design.
            * 2 option_name is empty.
            * 3 option_name is not found as key in Dict.
            * 4 option_sweep is empty, need at least one entry.
        """
        option_path = None
        a_value = None

        if len(option_sweep) == 0:
            return option_path, a_value, 4

        if option_name:
            option_path = option_name.split('.')
        else:
            return option_path, a_value, 2

        if qcomp_name in self.design.components.keys():
            qcomp_options = self.design.components[qcomp_name].options
        else:
            return option_path, a_value, 1

        a_value = qcomp_options

        # All but the last item in list.
        for name in option_path[:-1]:
            if name in a_value:
                a_value = self.option_value(a_value, name)
            else:
                self.design.logger.warning(f'Key="{name}" is not in dict.')
                return option_path, a_value, 3

        return option_path, a_value, 0

    @classmethod
    def option_value(cls, a_dict: Dict, search: str) -> str:
        """Get value from dict based on key.  This method is used for unknown
        depth, dict search, within a dict.

        Args:
            a_dict (Dict): Dictionary to get values from
            search (str): String to search for

        Returns:
            str: Value from the dictionary of the searched term.
        """
        value = a_dict[search]
        return value


# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Handles editing a QComponent."""

from PySide6 import QtCore, QtWidgets
from PySide6.QtCore import Qt, QTimer
from PySide6.QtWidgets import QTreeView, QAbstractItemView


class QTreeView_Base(QTreeView):
    """This class extends the `QTreeView` class."""

    def __init__(self, parent: QtWidgets.QWidget):
        """
        Args:
            parent (QtWidgets.QWidget): The widget
        """
        QTreeView.__init__(self, parent)
        QTimer.singleShot(200, self.style_me)
        self.expanded.connect(self.resize_on_expand)

    def style_me(self):
        """Style this widget."""
        # Can also do in the ui file, but doesn't always translate for some reason
        self.header().show()
        self.setAutoScroll(False)
        self.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)

    def autoresize_columns(self, max_width: int = 200):
        """Resize columns to content with maximum size.

        Args:
            max (int): Maximum window width.  Defaults to 200.
        """
        # For TreeView: resizeColumnToContents
        # For TableView: resizeColumnsToContents

        columns = self.model().columnCount(None)
        for i in range(columns):
            self.resizeColumnToContents(i)
            width = self.columnWidth(i)
            if width > max_width:
                self.setColumnWidth(i, max_width)

    def resize_on_expand(self):
        """Resize when exposed."""
        self.resizeColumnToContents(0)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""This is the CircleCaterpillar module."""

from qiskit_metal import draw, Dict  # , QComponent
from qiskit_metal.qlibrary.core import QComponent
#from qiskit_metal import is_true
from shapely.geometry import CAP_STYLE, JOIN_STYLE


class CircleRaster(QComponent):
    """A single configurable circle.

    Inherits QComponent class.

    .. image::
        CircleRaster.png

    .. meta::
        :description: Circle Raster

    Default Options:
        * radius: '300um'
        * resolution: '16'
        * cap_style: 'round' -- Valid options are 'round', 'flat', 'square'
        * subtract: 'False'
        * helper: 'False'
    """

    default_options = dict(
        radius='300um',
        resolution='16',
        cap_style='round',  # round, flat, square
        # join_style = 'round', # round, mitre, bevel
        # General
        subtract='False',
        helper='False')
    """Default drawing options"""

    TOOLTIP = """A single configurable circle"""

    def make(self):
        """The make function implements the logic that creates the geometry
        (poly, path, etc.) from the qcomponent.options dictionary of
        parameters, and the adds them to the design, using
        qcomponent.add_qgeometry(...), adding in extra needed information, such
        as layer, subtract, etc."""
        p = self.p  # p for parsed parameters. Access to the parsed options.

        # create the geometry
        circle = draw.Point(p.pos_x, p.pos_y).buffer(
            p.radius,
            resolution=int(p.resolution),
            cap_style=getattr(CAP_STYLE, p.cap_style),
            #join_style = getattr(JOIN_STYLE, p.join_style)
        )

        # add qgeometry
        self.add_qgeometry('poly', {'circle': circle},
                           subtract=p.subtract,
                           helper=p.helper,
                           layer=p.layer,
                           chip=p.chip)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable-msg=import-error
"""Reports a summary of information on Qiskit Metal and dependencies.

Contain functions to report more detailed information to orient a user,
used for debug purposes.
"""

from pathlib import Path
from typing import Union

import os
import sys
import getpass
import inspect
import platform
import webbrowser
import numpy
import qutip

from qiskit_metal.toolbox_python.display import Color, style_colon_list

__all__ = [
    'about', 'open_docs', 'orient_me', 'get_platform_info',
    'get_module_doc_page'
]


# pylint: disable-msg=invalid-name
# pylint: disable-msg=import-outside-toplevel
# pylint: disable-msg=bare-except
def about():
    """Reports a summary of information on Qiskit Metal and dependencies.

    Returns:
        str: About message
    """
    import qiskit_metal
    from PySide6.QtCore import __version__ as QT_VERSION_STR
    from PySide6 import __version__ as PYSIDE_VERSION_STR

    try:
        import matplotlib
        #matplotlib_ver = matplotlib.__version__
    except:
        #matplotlib_ver = 'None'
        pass

    try:
        from sip import SIP_VERSION_STR
    except:
        SIP_VERSION_STR = 'Not installed'
    # Riverbank: SIP is a tool for quickly writing Python modules that interface with
    # C++ and C libraries.

    text = f"""
Qiskit Metal        {qiskit_metal.__version__}

Basic
____________________________________
 Python              {sys.version}
 Platform            {platform.system()} {platform.machine()}
 Installation path   {os.path.dirname(inspect.getsourcefile(qiskit_metal))}

Packages
____________________________________
 Numpy               {numpy.__version__}
 Qutip               {qutip.__version__}

Rendering
____________________________________
 Matplotlib          {matplotlib.__version__}

GUI
____________________________________
 PySide6 version     {PYSIDE_VERSION_STR}
 Qt version          {QT_VERSION_STR}
 SIP version         {SIP_VERSION_STR}

IBM Quantum Team"""
    print(text)

    return text


######################################################################################
# DOCS


def get_module_doc_page(module,
                        folder=r'../docs/build/html',
                        page='index.html'):
    """Get the file path to a module doc folder assumed to be inside the
    package."""
    return Path(os.path.dirname(module.__file__)) / folder / page


def open_docs(page='https://qiskit-community.github.io/qiskit-metal/'):
    """Open the qiskit_metal documentation in HTML.

    Open the URL in new window, raising the window if possible.
    """
    webbrowser.open(page, new=1)

    ######################################################################################
    # More detailed information to orient a user.
    # For debug purposes.
    # Main function: ``orient_me```

    #def orient_me(do_print: bool = True) -> Union[None, str]:
    """Full system, python, user, and environemnt information.

    Args:
        do_print(bool): Return the string if True, else format and print.
    """


#    text = get_platform_info()
#    text += \
#        f" User and directories:\n\n"\
#        f"    User              : {getpass.getuser()}\n"\
#        f"    User home dirctry : {Path.home()}\n"\
#        f"    Current directory : {Path.cwd()}\n\n"\
#        f"    Conda default env : {os.environ.get('CONDA_DEFAULT_ENV', 'N/A')}\n"\
#        f"    Conda current env : {os.environ.get('CONDA_PREFIX', 'N/A')}\n"\
#       f"    Python executable : {sys.executable}\n"\

#    if do_print:
#        text = style_colon_list(text, Color.BOLD, Color.END)
#        print(text)
#        return None

#    return text


def get_platform_info() -> str:
    """Returns a string with the platform information."""

    return '''

 System platform information:

    system   : %s
    node     : %s
    release  : %s
    machine  : %s
    processor: %s
    summary  : %s
    version  : %s

 Python platform information:

    version  : %s (implem: %s)
    compiler : %s

''' % (platform.system(), platform.node(), platform.release(),
       platform.machine(), platform.processor(), platform.platform(),
       platform.version(), platform.python_version(),
       platform.python_implementation(), platform.python_compiler())



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.core import QComponent


class OpenToGround(QComponent):
    """A sample_shapes open to ground termination. Functions as a pin for auto drawing.

    Inherits `QComponent` class.

    .. image::
        OpenToGround.png

    .. meta::
        :description: Open to Ground

    Default Options:
        * width: '10um' -- The width of the 'cpw' terminating to ground (this is merely
          for the purpose of generating a value to pass to the pin)
        * gap: '6um' -- The gap of the 'cpw'
        * termination_gap: '6um' -- The length of dielectric from the end of the cpw center trace to the ground.

    Values (unless noted) are strings with units included, (e.g., '30um')
    """
    component_metadata = Dict(short_name='term', _qgeometry_table_poly='True')
    """Component metadata"""

    default_options = Dict(width='10um', gap='6um', termination_gap='6um')
    """Default connector options"""

    TOOLTIP = """A basic open to ground termination. """

    def make(self):
        """Build the component."""
        p = self.p  # p for parsed parameters. Access to the parsed options.

        port_line = draw.LineString([(0, -p.width / 2), (0, p.width / 2)])
        open_termination = draw.box(0, -(p.width / 2 + p.gap),
                                    p.termination_gap, (p.width / 2 + p.gap))
        # Rotates and translates the connector polygons (and temporary port_line)
        polys = [open_termination, port_line]
        polys = draw.rotate(polys, p.orientation, origin=(0, 0))
        polys = draw.translate(polys, p.pos_x, p.pos_y)
        [open_termination, port_line] = polys

        # Subtracts out ground plane on the layer its on
        self.add_qgeometry('poly', {'open_to_ground': open_termination},
                           subtract=True,
                           layer=p.layer)

        # Generates the pins
        self.add_pin('open', port_line.coords, p.width)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import shapely
import shapely.wkt as wkt

# Base Geometry Definitions - extend here
from shapely.geometry.base import BaseGeometry
from shapely.geometry import Point, LineString, Polygon
from shapely.geometry import box, shape
from shapely.ops import unary_union

from . import basic
from . import utility
from . import mpl

# Useful functions
from .utility import get_poly_pts, Vector
from .basic import rectangle, is_rectangle, flip_merge, rotate, translate, scale, buffer,\
    rotate_position, _iter_func_geom_, union, subtract
from .mpl import render, figure_spawn



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.core import QComponent
import numpy as np


class ShortToGround(QComponent):
    """A sample_shapes short to ground termination. Functions as a pin for auto CPW
    drawing.

    Inherits `QComponent` class.

    .. meta::
        :description: Short to Ground

    Default Options:
        * width: '10um' -- The width of the 'cpw' terminating to ground (this is merely for the purpose of
          generating a value to pass to the pin)

    Values (unless noted) are strings with units included, (e.g., '30um')
    """
    component_metadata = Dict(short_name='term')
    """Component metadata"""

    default_options = Dict(width='10um')
    """Default connector options"""

    TOOLTIP = """A basic short to ground termination"""

    def make(self):
        """Build the component."""
        p = self.p  # p for parsed parameters. Access to the parsed options.

        port_line = draw.LineString([(0, -p.width / 2), (0, p.width / 2)])

        # Rotates and translates the connector polygons (and temporary port_line)
        port_line = draw.rotate(port_line, p.orientation, origin=(0, 0))
        port_line = draw.translate(port_line, p.pos_x, p.pos_y)

        port_points = list(draw.shapely.geometry.shape(port_line).coords)

        #Generates the pin
        self.add_pin('short', port_points, p.width)



import numpy as np

from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.core import QComponent


class ReadoutResFC(QComponent):
    """Flip-chip readout resonator used in the flipchip tutorial.

    Geometry overview:
    - Circle centered at (``pos_x``, ``pos_y``) with radius ``readout_radius``.
    - Straight line (length ``readout_l1``) at 45 degrees.
    - 45 degree arc.
    - Vertical line (length ``readout_l2``).
    - 90 degree arc.
    - Horizontal line (length ``readout_l3``).
    - 180 degree arc.
    - Horizontal line (length ``readout_l4``).
    - Five meandering horizontal lines (length ``readout_l5``) separated by 180 degree arcs.

    The arc bend radius is ``readout_cpw_turnradius``, measured from the CPW center to the
    center of rotation. Lines and arcs form a CPW with width ``readout_cpw_width`` and
    gap ``readout_cpw_gap``.

    Tuning tips:
    - Change coupling to the qubit by varying ``readout_radius``.
    - Couple to the feedthrough line via the horizontal section of length ``readout_l3``.
    - Adjust resonator frequency by varying ``readout_l5``.
    """

    default_options = Dict(pos_x='0 um',
                           pos_y='0 um',
                           readout_radius='50 um',
                           readout_cpw_width='5 um',
                           readout_cpw_gap='5 um',
                           readout_cpw_turnradius='50 um',
                           readout_l1='400 um',
                           readout_l2='400 um',
                           readout_l3='400 um',
                           readout_l4='250 um',
                           readout_l5='400 um',
                           arc_step='1 um',
                           orientation='0',
                           layer='1',
                           layer_subtract='2',
                           subtract=True,
                           chip='main',
                           _default_connection_pads=Dict())
    ''' Default drawing options ?? '''
    component_metadata = Dict(short_name='readoutresfc',
                              _qgeometry_table_poly='True',
                              _qgeometry_table_junction='False')
    ''' Component metadata '''

    ##########################
    def make(self):
        ''' Make the head for readout res'''
        self.make_ro()

    ##########################

    def make_ro(self):
        """Create the head of the readout resonator.

        Contains the circular coupling patch, 45 deg line, 45 deg arc, and a short
        straight segment (of length w) for smooth subtraction.
        """
        # access to parsed values from the user option
        p = self.p

        # access to chip name
        chip = p.chip

        # local variables
        r = p.readout_radius
        w = p.readout_cpw_width
        g = p.readout_cpw_gap
        turnradius = p.readout_cpw_turnradius
        l_1 = p.readout_l1
        l_2 = p.readout_l2
        l_3 = p.readout_l3
        l_4 = p.readout_l4
        l_5 = p.readout_l5

        # create the coupling patch in term of a circle
        cppatch = draw.Point(0, 0).buffer(r)

        # create the extended arm
        ## useful coordinates
        x_1, y_1 = l_1 * np.cos(np.pi / 4), -l_1 * np.sin(np.pi / 4)
        x_2, y_2 = x_1 + turnradius * (
            1 - np.cos(np.pi / 4)), y_1 - turnradius * np.sin(np.pi / 4)
        coord_init = draw.Point(x_1, y_1)
        coord_center = draw.Point(x_1 - turnradius * np.cos(np.pi / 4),
                                  y_1 - turnradius * np.sin(np.pi / 4))
        x_3, y_3 = x_2, y_2
        x_4, y_4 = x_3, y_3 - l_2
        x_5, y_5 = x_4 + turnradius, y_4
        coord_init1 = draw.Point(x_4, y_4)
        coord_center1 = draw.Point(x_5, y_5)
        x_6, y_6 = x_5, y_5 - turnradius
        x_7, y_7 = x_5 + l_3, y_6
        x_8, y_8 = x_7, y_7 + turnradius
        coord_init2 = draw.Point((x_7, y_7))
        coord_center2 = draw.Point((x_8, y_8))
        x_9, y_9 = x_8, y_8 + turnradius
        x_10, y_10 = x_8 - l_4, y_9
        x_11, y_11 = x_10, y_10 + turnradius
        coord_init3 = draw.Point((x_10, y_10))
        coord_center3 = draw.Point((x_11, y_11))
        arc3 = self.arc(coord_init3, coord_center3, -np.pi)
        x_12, y_12 = x_11, y_11 + turnradius
        x_13, y_13 = x_12 + l_5, y_12
        line12 = draw.LineString([(x_12, y_12), (x_13, y_13)])
        x_14, y_14 = x_13, y_13 + turnradius
        coord_init4 = draw.Point((x_13, y_13))
        coord_center4 = draw.Point((x_14, y_14))
        arc4 = self.arc(coord_init4, coord_center4, np.pi)
        ## line containing the 45deg line, 45 deg arc,
        ## and a short straight segment for smooth subtraction
        cparm_line = draw.shapely.ops.unary_union([
            draw.LineString([(0, 0), coord_init]),
            self.arc(coord_init, coord_center, -np.pi / 4),
            draw.LineString([(x_3, y_3), (x_4, y_4)]),
            self.arc(coord_init1, coord_center1, np.pi / 2),
            draw.LineString([(x_6, y_6), (x_7, y_7)]),
            self.arc(coord_init2, coord_center2, np.pi),
            draw.LineString([(x_9, y_9), (x_10, y_10)]), arc3, line12, arc4,
            draw.translate(line12, 0, 2 * turnradius),
            draw.translate(arc3, 0, 4 * turnradius),
            draw.translate(line12, 0, 4 * turnradius),
            draw.translate(arc4, 0, 4 * turnradius),
            draw.translate(line12, 0, 6 * turnradius),
            draw.translate(arc3, 0, 8 * turnradius),
            draw.translate(line12, 0, 8 * turnradius)
        ])
        cparm = cparm_line.buffer(w / 2, cap_style=2, join_style=1)
        ## fix the gap resulting from buffer
        eps = 1e-3
        cparm = draw.Polygon(cparm.exterior)
        cparm = cparm.buffer(eps, join_style=2).buffer(-eps, join_style=2)

        # create combined objects for the signal line and the etch
        ro = draw.shapely.ops.unary_union([cppatch, cparm])
        ro_etch = ro.buffer(g, cap_style=2, join_style=2)
        x_15, y_15 = x_14, y_14 + 7 * turnradius
        x_16, y_16 = x_15 + g / 2, y_15
        port_line = draw.LineString([(x_15, y_15 + w / 2),
                                     (x_15, y_15 - w / 2)])
        subtract_patch = draw.LineString([(x_16, y_16 - w / 2 - g - eps),
                                          (x_16, y_16 + w / 2 + g + eps)
                                         ]).buffer(g / 2, cap_style=2)
        ro_etch = ro_etch.difference(subtract_patch)

        # rotate and translate
        polys = [ro, ro_etch, port_line]
        polys = draw.rotate(polys, p.orientation, origin=(0, 0))
        polys = draw.translate(polys, p.pos_x, p.pos_y)

        # update each object
        [ro, ro_etch, port_line] = polys

        # generate QGeometry
        self.add_qgeometry('poly', dict(ro=ro), chip=chip, layer=p.layer)
        self.add_qgeometry('poly',
                           dict(ro_etch=ro_etch),
                           chip=chip,
                           layer=p.layer_subtract,
                           subtract=p.subtract)

        # generate pins
        self.add_pin('readout', port_line.coords, width=w, gap=g, chip=chip)

    def arc(self, coord_init, coord_center, angle):
        '''
        Generate x,y coordinates (in terms of shapely.geometry.Point()) of an arc with:
        a specified initial point, rotation center, and
        rotation direction (specified by angle in radian (float or integer), positive is ccw).

        coord_init, and coord_center should be shapely.geometry.Point object
        '''
        # access to parse values from the user option
        p = self.p

        # local variable
        r = p.readout_cpw_turnradius
        step = p.arc_step

        # determine step number
        step_angle = step / r if angle >= 0 else -step / r
        step_N = abs(int(angle / step_angle))
        #laststep_flag = True if angle % step_angle != 0 else False
        laststep_flag = bool(angle % step_angle != 0)

        # generate coordinate
        coord = [coord_init]
        point = coord_init
        for i in range(step_N):
            point = draw.rotate(point,
                                step_angle,
                                origin=coord_center,
                                use_radians=True)
            coord.append(point)
        if laststep_flag:
            point = draw.rotate(coord_init,
                                angle,
                                origin=coord_center,
                                use_radians=True)
            coord.append(point)
        coord = draw.LineString(coord)
        return coord
